#include "defines.h"
#include "kozos.h"
#include "intr.h"
#include "interrupt.h"
#include "lib.h"
#include "ether_wrap.h"
#include "ethernet.h"
#include "netdrv.h"

static unsigned char my_macaddr[MACADDR_SIZE];

static struct netbuf *send_buf_head;
static struct netbuf **send_buf_tail = &send_buf_head;

static void recv_packet()
{
  struct netbuf *pkt;

  while (1) {
    pkt = kz_kmalloc(DEFAULT_NETBUF_SIZE);
    memset(pkt, 0, DEFAULT_NETBUF_SIZE);
    pkt->cmd  = ETHERNET_CMD_RECV;

    /*
     * ethernetե졼ब14ХȤǣܿǤʤΤǡethernetե졼ʹߤ
     * Хȥ饤Ȥ褦ˡǡƬ򣲥Хȶ롥
     */
    pkt->top  = pkt->data + 2;

    pkt->size = ether_recv(0, pkt->top);
    if (pkt->size > 0) {
      kz_send(MSGBOX_ID_ETHPROC, 0, (char *)pkt);
    } else {
      kz_kmfree(pkt);
      break;
    }
  }
}

static void send_packet(struct netbuf *buf)
{
  /* üɲ */
  if (buf) {
    buf->next = NULL;
    *send_buf_tail = buf;
    send_buf_tail = &buf->next;
  }

  if (send_buf_head) {
    if (!buf || !ether_intr_is_send_enable(0)) {
      /* ƬȴФ */
      buf = send_buf_head;
      send_buf_head = send_buf_head->next;
      if (!send_buf_head)
	send_buf_tail = &send_buf_head;
      buf->next = NULL;

      /*
       * ̵ʤСϤƤʤΤϤ롥
       * ͭʤϤƤꡤߤαĹ
       * ХåեΥǡ缡Τǡ⤷ʤƤ褤
       */
      if (!ether_intr_is_send_enable(0)) {
	ether_intr_send_enable(0); /* ͭ */
      }

      ether_send(0, buf->size, buf->top);
      kz_kmfree(buf);
    }
  }
}

/* ߥϥɥ */
static void netdrv_intr(void)
{
  struct netbuf *pkt;

  ether_intr_clear(0);

  if (ether_is_recv_enable(0)) {
    ether_intr_clear_recv(0);
    pkt = kx_kmalloc(DEFAULT_NETBUF_SIZE);
    memset(pkt, 0, DEFAULT_NETBUF_SIZE);
    pkt->cmd  = NETDRV_CMD_RECVINTR;
    kx_send(MSGBOX_ID_NETPROC, 0, (char *)pkt);
  }

  if (ether_is_send_enable(0)) {
    ether_intr_clear_send(0);
    pkt = kx_kmalloc(DEFAULT_NETBUF_SIZE);
    memset(pkt, 0, DEFAULT_NETBUF_SIZE);
    pkt->cmd  = NETDRV_CMD_SENDINTR;
    kx_send(MSGBOX_ID_NETPROC, 0, (char *)pkt);
  }
}

static int netdrv_init(void)
{
  ether_init(0, my_macaddr);
  return 0;
}

/* åɤ׵ */
static int netdrv_proc(struct netbuf *buf)
{
  int ret = 0;

  switch (buf->cmd) {
  case NETDRV_CMD_USE: /* ͥåȡɥ饤Фλѳ */
    /* ether_init(0, my_macaddr); */ /* netdrv_init()ǹԤäƤΤ */
    ether_intr_recv_enable(0); /* ͭ() */
    buf = kz_kmalloc(sizeof(*buf));
    buf->cmd = ETHERNET_CMD_MACADDR;
    memcpy(buf->option.common.macaddr.addr, my_macaddr, MACADDR_SIZE);
    kz_send(MSGBOX_ID_ETHPROC, 0, (char *)buf);
    break;

  case NETDRV_CMD_SEND: /* ͥåȤؤΥե졼 */
    /* ػߤˤɬפ̵ */
    send_packet(buf);
    ret = 1;
    break;

  case NETDRV_CMD_RECVINTR: /* ͥåȤΥե졼 */
    /* ػߤˤɬפ̵ */
    recv_packet();
    break;

  case NETDRV_CMD_SENDINTR: /* ͥåȤؤΥե졼ϴλ */
    /* ػߤˤɬפ̵ */
    if (!send_buf_head) {
      /* ǡ̵ʤСλ */
      ether_intr_send_disable(0);
    } else {
      /* ǡʤС³ */
      send_packet(NULL);
    }
    break;

  default:
    break;
  }

  return ret;
}

int netdrv_main(int argc, char *argv[])
{
  struct netbuf *buf;
  int ret;
  static int initialized = 0;

  if (!initialized) {
    netdrv_init();
    kz_setintr(SOFTVEC_TYPE_ETHINTR, netdrv_intr); /* ߥϥɥ */

    initialized++;
  }

  if (kz_recv(MSGBOX_ID_NETPROC, NULL, (void *)&buf) != -1) {
    ret = netdrv_proc(buf);
    if (!ret) kz_kmfree(buf);
  }

  return 0;
}
